# Copyright (c) 2024, Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

from .g2p_eng import G2pEn
from pypinyin.constants import RE_HANS


g2p_en = G2pEn()


class Token:
    def __init__(self, word, pos=None, phones=None):
        self.word = word
        self.pos = pos
        self.phones = phones if phones else word
        if RE_HANS.match(word):
            self.lang = "ZH"
        elif word.isdigit():
            self.lang = "NUM"
        elif word.replace("'", "").isalnum():
            self.phones = [phone for phone in g2p_en.g2p(self.word) if phone != " "]
            self.lang = "EN"
        else:
            self.lang = "SYM"

    def __getitem__(self, item):
        return self.__dict__[item]

    def tone(self, index):
        if self.lang != "ZH" or index >= len(self.phones):
            return -1
        return self.phones[index][-1]

    def sandhi(self, tone, index=0):
        assert index < len(self.phones)
        self.phones[index] = self.phones[index][:-1] + str(tone)

    def push_front(self, token):
        self.word = token.word + self.word
        self.phones = token.phones + self.phones
        return self

    def push_back(self, token):
        self.word = self.word + token.word
        self.phones = self.phones + token.phones
        return self

    def __str__(self):
        return json.dumps(
            {k: v for k, v in self.__dict__.items() if k != "pos"}, ensure_ascii=False
        )

    def __repr__(self):
        return self.__str__()



# Copyright (c) 2021 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jieba

from .token import Token


class ToneSandhi:
    def __init__(self):
        self.digits = list("零一二三四五六七八九十")
        self.neural_tone_words = (
            "麻烦 麻利 鸳鸯 高粱 骨头 骆驼 马虎 首饰 馒头 馄饨 风筝 难为 队伍 阔气 闺女 门道 锄头 铺盖 铃铛 铁匠 钥匙 里脊 里头 部分 "
            "那么 道士 造化 迷糊 连累 这么 这个 运气 过去 软和 转悠 踏实 跳蚤 跟头 趔趄 财主 豆腐 讲究 记性 记号 认识 规矩 见识 裁缝 "
            "补丁 衣裳 衣服 衙门 街坊 行李 行当 蛤蟆 蘑菇 薄荷 葫芦 葡萄 萝卜 荸荠 苗条 苗头 苍蝇 芝麻 舒服 舒坦 舌头 自在 膏药 脾气 "
            "脑袋 脊梁 能耐 胳膊 胭脂 胡萝 胡琴 胡同 聪明 耽误 耽搁 耷拉 耳朵 老爷 老实 老婆 戏弄 将军 翻腾 罗嗦 罐头 编辑 结实 红火 "
            "累赘 糨糊 糊涂 精神 粮食 簸箕 篱笆 算计 算盘 答应 笤帚 笑语 笑话 窟窿 窝囊 窗户 稳当 稀罕 称呼 秧歌 秀气 秀才 福气 祖宗 "
            "砚台 码头 石榴 石头 石匠 知识 眼睛 眯缝 眨巴 眉毛 相声 盘算 白净 痢疾 痛快 疟疾 疙瘩 疏忽 畜生 生意 甘蔗 琵琶 琢磨 琉璃 "
            "玻璃 玫瑰 玄乎 狐狸 状元 特务 牲口 牙碜 牌楼 爽快 爱人 热闹 烧饼 烟筒 烂糊 点心 炊帚 灯笼 火候 漂亮 滑溜 溜达 温和 清楚 "
            "消息 浪头 活泼 比方 正经 欺负 模糊 槟榔 棺材 棒槌 棉花 核桃 栅栏 柴火 架势 枕头 枇杷 机灵 本事 木头 木匠 朋友 月饼 月亮 "
            "暖和 明白 时候 新鲜 故事 收拾 收成 提防 挖苦 挑剔 指甲 指头 拾掇 拳头 拨弄 招牌 招呼 抬举 护士 折腾 扫帚 打量 打算 打扮 "
            "打听 打发 扎实 扁担 戒指 懒得 意识 意思 悟性 怪物 思量 怎么 念头 念叨 别人 快活 忙活 志气 心思 得罪 张罗 弟兄 开通 应酬 "
            "庄稼 干事 帮手 帐篷 希罕 师父 师傅 巴结 巴掌 差事 工夫 岁数 屁股 尾巴 少爷 小气 小伙 将就 对头 对付 寡妇 家伙 客气 实在 "
            "官司 学问 字号 嫁妆 媳妇 媒人 婆家 娘家 委屈 姑娘 姐夫 妯娌 妥当 妖精 奴才 女婿 头发 太阳 大爷 大方 大意 大夫 多少 多么 "
            "外甥 壮实 地道 地方 在乎 困难 嘴巴 嘱咐 嘟囔 嘀咕 喜欢 喇嘛 喇叭 商量 唾沫 哑巴 哈欠 哆嗦 咳嗽 和尚 告诉 告示 含糊 吓唬 "
            "后头 名字 名堂 合同 吆喝 叫唤 口袋 厚道 厉害 千斤 包袱 包涵 匀称 勤快 动静 动弹 功夫 力气 前头 刺猬 刺激 别扭 利落 利索 "
            "利害 分析 出息 凑合 凉快 冷战 冤枉 冒失 养活 关系 先生 兄弟 便宜 使唤 佩服 作坊 体面 位置 似的 伙计 休息 什么 人家 亲戚 "
            "亲家 交情 云彩 事情 买卖 主意 丫头 丧气 两口 东西 东家 世故 不由 下水 下巴 上头 上司 丈夫 丈人 一辈 那个 菩萨 父亲 母亲 "
            "咕噜 邋遢 费用 冤家 甜头 介绍 荒唐 大人 泥鳅 幸福 熟悉 计划 扑腾 蜡烛 姥爷 照顾 喉咙 吉他 弄堂 蚂蚱 凤凰 拖沓 寒碜 糟蹋 "
            "倒腾 报复 逻辑 盘缠 喽啰 牢骚 咖喱 扫把 惦记 上来 下来 进来 出来 回来 过来 起来 上去 下去 进去 出去 回去 过去 几个 有个 "
            "两个 半个 多个 各个 整个 每个 做个 是个 房子 车子 胖子 瘦子 矮子 垫子 饺子 妻子 儿子 桌子 椅子 嫂子 屋子 面子 样子 盘子 "
            "瓶子 脑子 旗子 乱子 乐子 小子 法子 票子 圈子 调子 场子 位子 瞎子 镜子 方子 耗子 帽子 爷爷 奶奶 试试 姐姐 哥哥 弟弟 妹妹 "
            "爸爸 妈妈 姥姥 舅舅 婶婶 嫂嫂 宝宝"
        ).split()
        self.not_neural_tone_words = (
            "女儿 妻儿 胎儿 婴儿 幼儿 少儿 小儿 孤儿 新生儿 婴幼儿"
        ).split()

    def _merge(self, tokens):
        new_tokens = []
        idx = 0
        while idx < len(tokens):
            token = tokens[idx]
            idx += 1
            if not new_tokens or new_tokens[-1].lang != "ZH" or token.lang != "ZH":
                new_tokens.append(token)
                continue
            last_token = new_tokens[-1]
            # 1. merge single 不, 很, 一 and the word behind it.
            # 2. merge single 儿 and the word before it.
            if last_token.word in ("不", "很", "一") or token.word in ("儿"):
                last_token.push_back(token)
            # 3. merge continuous three tones, e.g. 小/火车 => 小火车, 也/很/好 => 也很/好 => 也很好
            elif (
                last_token.tone(-1) == "3"
                and token.tone(0) == "3"
                and len(last_token.word) <= 2
                and len(token.word) <= 2
                and len(last_token.word) + len(token.word) <= 4
            ):
                last_token.push_back(token)
            # 4. merge 一 and reduplication verbs around it, e.g. 听/一/听 => 听一听
            elif (
                token.word == "一"
                and last_token.pos[0] == "v"
                and idx < len(tokens)
                and last_token.word == tokens[idx].word
            ):
                last_token.push_back(token).push_back(tokens[idx])
                idx += 1
            else:
                new_tokens.append(token)
        return new_tokens

    def _bu_sandhi(self, token):
        # e.g. 看不懂, 去不去
        if len(token.word) == 3 and token.word[1] == "不":
            token.sandhi("5", 1)
        else:
            for i, char in enumerate(token.word):
                # 不 before tone4 should be bu2, e.g. 不怕
                if (
                    char == "不"
                    and i + 1 < len(token.word)
                    and token.tone(i + 1) == "4"
                ):
                    token.sandhi("2", i)

    def _yi_sandhi(self, token):
        # 一 between reduplication words should be yi5, e.g. 看一看
        if (
            len(token.word) == 3
            and token.word[1] == "一"
            and token.word[0] == token.word[-1]
        ):
            token.sandhi("5", 1)
        else:
            for i, char in enumerate(token.word):
                if char != "一":
                    continue
                prev_char = token.word[i - 1] if i > 0 else ""
                next_char = token.word[i + 1] if i + 1 < len(token.word) else ""
                if prev_char in self.digits + ["万", "月", "第", "初"]:
                    token.sandhi("1", i)
                # 无法判断是否为序数时，当成序数不变调处理, e.g. 一月 => yi1 yue4
                elif next_char in ["", "月", "班", "连", "楼"]:
                    token.sandhi("1", i)
                # 一 before tone4 should be yi2, e.g. 一段
                elif token.tone(i + 1) == "4":
                    token.sandhi("2", i)
                # 一 before non-tone4 should be yi4, e.g. 一天
                else:
                    token.sandhi("4", i)

    # finer split a long word into twos for neural & three tone sandhi
    # e.g. 很漂亮 => 很/漂亮, 与好姐妹 => 与/好姐妹
    @staticmethod
    def _split_word(word):
        word_list = jieba.cut_for_search(word)
        word_list = sorted(word_list, key=len, reverse=False)
        first_subword = word_list[0]
        first_begin_idx = word.find(first_subword)
        if first_begin_idx == 0:
            second_subword = word[len(first_subword) :]
            new_word_list = [first_subword, second_subword]
        else:
            second_subword = word[: -len(first_subword)]
            new_word_list = [second_subword, first_subword]
        return new_word_list

    def _neural_sandhi(self, token):
        idx = 0
        for sub_word in self._split_word(token.word):
            idx += len(sub_word)
            if not sub_word or sub_word in self.not_neural_tone_words:
                continue
            if sub_word in self.neural_tone_words:
                token.sandhi("5", idx - 1)
            # 儿话音、们字变调
            if sub_word[-1] == ["儿", "们"]:
                token.sandhi("5", idx - 1)
            # 语气词变调
            if (
                len(sub_word) >= 1
                and sub_word[-1] in "吧呢啊呐噻嘛吖嗨呐哦哒滴哩哟喽啰耶喔诶"
            ):
                token.sandhi("5", idx - 1)
            if sub_word in "了着过" and token.pos in {"n", "v", "a"}:
                token.sandhi("5", idx - 1)
        return token

    def _all_tone_three(self, pinyins):
        return all(pinyin.endswith("3") for pinyin in pinyins)

    def _three_sandhi(self, token):
        if len(token.word) == 2:
            if self._all_tone_three(token.phones):
                token.sandhi("2", 0)
        elif len(token.word) >= 3:
            sub_words = self._split_word(token.word)
            if len(sub_words[0]) == 1:
                # e.g. 纸/老虎、与/好姐妹
                if self._all_tone_three(token.phones[1:3]):
                    token.sandhi("2", 1)
                # e.g. 好/喜欢
                elif self._all_tone_three(token.phones[:2]):
                    token.sandhi("2", 0)
            else:
                # e.g. 广场/舞、字母/表、所有/人
                if self._all_tone_three(token.phones[:2]):
                    token.sandhi("2", 0)
                if self._all_tone_three(token.phones[1:3]):
                    # e.g. 坎坎/坷坷 => kan2 kan3 ke2 ke3
                    if len(token.word) <= 3 or not self._all_tone_three(
                        token.phones[2:4]
                    ):
                        token.sandhi("2", 1)
                # e.g. 期待/已久、省委/党校
                if len(token.word) > 3 and self._all_tone_three(token.phones[2:4]):
                    token.sandhi("2", 2)

    def modified_tone(self, tokens):
        tokens = self._merge(tokens)
        for token in tokens:
            if len(token.word) < 2 or token.lang != "ZH":
                continue
            self._bu_sandhi(token)
            self._yi_sandhi(token)
            self._neural_sandhi(token)
            self._three_sandhi(token)
        return tokens



# Copyright (c) 2024, Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# 5
TONES = {"0", "1", "2", "3", "4"}

# 21
# pypinyin to_initials(strict=True)
# 对于零声母拼音(没有声母)，声母为空字符串
STRICT_INITIALS = set("b p m f d t n l g k h j q x zh ch sh r z s c".split()).union("")

# 严格意义来说，y 和 w 不算声母。对于零声母拼音（没有声母）：
# 1. 韵母 i 后面有其他韵母，把 i 换成 y
# 2. 韵母 u 后面有其他韵母，把 u 换成 w
# 3. 韵母 i 后面没有其他韵母，i 前面加 y
# 4. 韵母 u 后面没有其他韵母，u 前面加 w
# 5. 韵母 ü 后面有其他韵母，去掉两点写成 u，且前面加 y。如 yue
# pypinyin to_initials(strict=False)
INITIALS = set("b p m f d t n l g k h j q x zh ch sh r z s c".split()).union({"y", "w"})

# 3 个 后鼻音韵母：n, ng, m。替换 n 和 m 为 ng 和 mg，避免与声母混淆
POSTNASALS = {
    "n": "ng",
    "m": "mg",
}

# 39 个韵母：
# a o e i -i(前) -i(后) u ü er ai ei ao ou ia ie ua uo
# üe iao iou uai uei an ian uan üan en in uen ün ang
# iang uang eng ing ueng ong iong ê
# 舌尖前元音 -i(前) 舌尖要靠前，是前高不圆唇舌尖元音，只能与舌尖前音 z c s 构成整体认读音节
# 舌尖后元音 -i(后) 舌尖要靠后，是后高不圆唇舌尖元音，只能与舌尖后音 zh ch sh r 后构成整体认读音节
# ê 只在语气词"欸"(现用"唉")中单用

# 37
# pypinyin to_finals(strict=False)
# - 不区分 i, -i(前) 和 -i(后)，统一成 i
# - 少了 uen 和 ueng (只出现在零声母拼音中，会写成 wen 和 weng)
# - 少了 üan (在拼音中均写成 uan)
# - üe 多了一种写法 ue (根据《汉语拼音方案》，ü 跟 n, l 以外的声母相拼时，写成 u)
# - 简写：iou => iu, uei => ui, ün => un
STRICT_FINALS = set(
    (
        "a o e i u v er ai ei ao ou ia ie ua uo "
        "ve iao iu uai ui an ian uan en in ang un "
        "iang uang eng ing ong iong ê ng mg ue"
    ).split()
)

# 40
# pypinyin to_finals(strict=True)
# - 不区分 i, -i(前) 和 -i(后)，统一成 i
# - 多了 io (《现代汉语词典》中，“哟”读 io1)
FINALS = set(
    (
        "a o e i u v er ai ei ao ou ia ie ua uo "
        "ve iao iou uai uei an ian uan van en in uen vn ang "
        "iang uang eng ing ueng ong iong ê ng mg io"
    ).split()
)

# http://www.speech.cs.cmu.edu/cgi-bin/cmudict
STRESS = {"", "0", "1", "2"}

# 39
PHONES = set(
    (
        "AA AE AH AO AW AY B CH D DH EH ER EY F G HH IH IY "
        "JH K L M N NG OW OY P R S SH T TH UH UW V W Y Z ZH"
    ).split()
)



# Copyright (c) 2024, Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import nltk
from nltk.corpus import cmudict
import wordsegment

dirname = os.path.dirname(__file__)
nltk.data.path.insert(0, f"{dirname}/nltk_data")

import g2p_en


class G2pEn:
    def __init__(self):
        wordsegment.load()
        self.g2p_e = g2p_en.G2p()
        self.cmudict = cmudict.dict()
        # Remove abbreviations badcase like "HUD" in cmudict.
        for word in ["AE", "AI", "AR", "IOS", "HUD", "OS"]:
            del self.cmudict[word.lower()]

    def g2p_ch(self, ch):
        assert len(ch) == 1
        # In abbreviations, "A" should be pronounced as "EY1", not "AH0".
        if ch.upper() == "A":
            return ["EY1"]
        return self.g2p_e(ch)

    def g2p_abbr(self, word):
        phones = []
        for ch in word:
            phones.extend(self.g2p_ch(ch))
        return phones

    def g2p(self, word):
        # 大写单词长度小于等于 3，按字母念
        if word.isupper() and len(word) <= 3:
            # e.g. "IT" => "I T"
            return self.g2p_abbr(word)
        # 单词在 CMU dict 中，按照 CMU dict 念
        if word.lower() in self.cmudict:
            return self.cmudict[word.lower()][0]
        # 小写 oov 长度小于等于 3，大小 oov 长度小于等于 4，按字母念
        # e.g. tts => t t s, WFST => W F S T
        if (word.islower() and len(word) <= 3) or (word.isupper() and len(word) <= 4):
            return self.g2p_abbr(word)
        # 其他 OOV 先转小写，用 wordsegment 分词
        # e.g. wifi => wifi, autojs => auto js
        bpes = wordsegment.segment(word.lower())
        # 无法分词则让 g2p_en 预测
        if len(bpes) == 1:
            return self.g2p_e(word)
        # 分词结果递归做 g2p
        phones = []
        for bpe in bpes:
            phones.extend(self.g2p_ch(bpe) if len(bpe) == 1 else self.g2p(bpe))
        return phones



# Copyright (c) 2023 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from setuptools import setup, find_packages


with open("README.md", encoding="utf8") as fin:
    long_description = fin.read()

with open("requirements.txt", encoding="utf-8") as f:
    requirements = f.readlines()
extras_require = {"g2pw": ["torch", "modelscope", "pypinyin-g2pw"]}

setup(
    name="g2p-mix",
    version=os.getenv("BUILD_VERSION") or "0.0.1",
    author="Zhendong Peng",
    author_email="pzd17@tsinghua.org.cn",
    long_description=long_description,
    long_description_content_type="text/markdown",
    description="G2P mix",
    url="https://github.com/pengzhendong/g2p-mix",
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements,
    extras_require=extras_require,
    entry_points={
        "console_scripts": [
            "g2p_mix = g2p_mix.cli:main",
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
    ],
)



# Copyright (c) 2024, Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re

from jieba import posseg
from pypinyin import lazy_pinyin, Style
from pypinyin.constants import RE_HANS
from pypinyin.contrib.tone_convert import to_initials, to_finals

from .constants import POSTNASALS
from .token import Token
from .tone_sandhi import ToneSandhi


os.environ["TRANSFORMERS_OFFLINE"] = "1"
os.environ["HF_DATASETS_OFFLINE"] = "1"


class G2pMix:
    def __init__(self, use_g2pw=False, repo_id=None, model_dir=None, model_source=None):
        if use_g2pw:
            from modelscope.hub.snapshot_download import snapshot_download
            from pypinyin_g2pw import G2PWPinyin

            if not model_dir or not model_source:
                repo_id = repo_id or "pengzhendong/g2pw"
                repo_dir = snapshot_download(repo_id)
                model_dir = f"{repo_dir}/G2PWModel"
                model_source = f"{repo_dir}/bert-base-chinese"
            self.lazy_pinyin = G2PWPinyin(
                model_dir=model_dir,
                model_source=model_source,
                neutral_tone_with_five=True,
            ).lazy_pinyin
        else:
            self.lazy_pinyin = lazy_pinyin
        self.sandhier = ToneSandhi()
        # add space between ascii and chinese characters
        self.pattern = re.compile(
            r"(?<=[\u4e00-\u9fa5])(?=[\x00-\x19\x21-\x7F])|(?<=[\x00-\x19\x21-\x7F])(?=[\u4e00-\u9fa5])"
        )
        # load dict to fix some badcase of pypinyin
        self.load_dict()

    def load_dict(self):
        # 为、曾、更、长
        # from pypinyin.constants import PINYIN_DICT
        # print(hex(ord("为")), PINYIN_DICT[ord("为")])
        from pypinyin import load_single_dict
        load_single_dict({ord("长"): "cha2ng,zha4ng"}, "tone2")
        load_single_dict({ord("为"): "we4i,we2i"}, "tone2")
        # from pypinyin import load_phrases_dict
        # load_phrases_dict({"长时间": [["cha2ng"], ["shi2"], ["jia1n"]]}, "tone2")
        # from pypinyin_dict.pinyin_data import zdic
        # zdic.load()  # 汉典

    def get_pinyins(self, text):
        segs = list(posseg.cut(text))
        words = [word for word, _ in segs]
        # tone_sandhi rules of pypinyin is not perfect
        pinyins = self.lazy_pinyin(words, neutral_tone_with_five=True, style=Style.TONE3)

        # combine pinyins, english words and punctuations by `posseg.cut`'s output
        idx = 0
        tokens = []
        last_word = " "
        for i, (word, pos) in enumerate(segs):
            next_word = words[i + 1] if i + 1 < len(words) else " "
            if word == " ":
                idx += 1
            elif RE_HANS.match(word):
                tokens.append(Token(word, pos, pinyins[idx : idx + len(word)]))
                idx += len(word)
            # he' ve => he've, ' cause => ' cause
            elif (
                (word.isalnum() and len(last_word) > 1 and last_word[-1] == "'")
                or (word[0] == "'" and len(word) > 1 and last_word != " ")
                or (word == "'" and last_word != " " and next_word != " ")
            ):
                word = tokens[-1].word + word
                tokens[-1] = Token(word, pos)
                idx += 1
            else:
                tokens.append(Token(word, pos))
                idx += 1
            last_word = word
        return tokens

    def g2p(self, text, strict=False, sandhi=False, return_seg=False):
        text = self.pattern.sub(" ", text)
        seg_tokens = self.get_pinyins(text)
        if sandhi:
            seg_tokens = self.sandhier.modified_tone(seg_tokens)
        tokens = []
        for token in seg_tokens:
            if token.lang != "ZH":
                tokens.append(token)
                continue
            # split pinyin into initial and final
            for i, pinyin in enumerate(token.phones):
                tone = pinyin[-1]
                if pinyin[:-1] in POSTNASALS:
                    token.phones[i] = ["", POSTNASALS[pinyin[:-1]] + tone]
                else:
                    initial = to_initials(pinyin, strict=strict)
                    final = to_finals(pinyin, strict=strict) + tone
                    token.phones[i] = [initial, final]
                tokens.append(Token(token.word[i], token.pos, token.phones[i]))
        # 中文保留分词结果 or 中文分词结果拆成单个汉字
        return tokens if not return_seg else seg_tokens
